use std::collections::BTreeSet;
use std::fs::File;
use std::io::{self, BufRead, BufReader};
use std::iter::{repeat, successors};

fn next(c: (i32, i32), direction: &str) -> (i32, i32) {
    match direction {
        "R" => (c.0 + 1, c.1),
        "L" => (c.0 - 1, c.1),
        "U" => (c.0, c.1 + 1),
        "D" => (c.0, c.1 - 1),
        _ => panic!("Wrong direction format!"),
    }
}

fn direction_length(entry: String) -> (String, usize) {
    /// Creates a (direction, length) pair to later produce a string
    /// on a direction with the given length. E.g. given "R10" the output of
    (
        entry.chars().nth(0).unwrap().to_string(),
        entry[1..].parse().unwrap(),
    )
}

fn to_coordinates(entries: Vec<String>) -> BTreeSet<(i32, i32)> {
    let mut wire = BTreeSet::new();

    let mut it = entries
        .into_iter()
        .map(direction_length)
        .flat_map(|(direction, length)| repeat(direction).take(length));

    successors(
        it.next()
            .map(|direction| (next((0, 0), direction.as_ref()))),
        |c| it.next().map(|direction| (next(*c, direction.as_ref()))),
    )
    .for_each(|c| {
        wire.insert(c);
    });

    wire
}

fn distance(position: (i32, i32)) -> i32 {
    position.0.abs() + position.1.abs()
}

pub fn main() -> io::Result<()> {
    let f = BufReader::new(File::open("day3-input.txt")?);
    let entries: Vec<BTreeSet<(i32, i32)>> = f
        .lines()
        .map(|line| line.unwrap().split(",").map(String::from).collect())
        .map(to_coordinates)
        .collect();

    let closest: Option<i32> = entries[0]
        .intersection(&entries[1])
        .cloned()
        .map(distance)
        .min();
    println!("{:?}", closest);

    Ok(())
}

#[test]
fn test_coordinates_for_entry() {
    let mut expected: BTreeSet<(i32, i32)> = BTreeSet::new();
    for i in 1..11 {
        expected.insert((i, 0));
    }
    assert_eq!(to_coordinates(vec!["R10".to_owned()]), expected);

    expected = BTreeSet::new();
    expected.insert((1, 0));
    expected.insert((2, 0));
    expected.insert((2, 1));
    expected.insert((2, 2));
    let input = "R2,U2".to_owned().split(",").map(String::from).collect();
    assert_eq!(to_coordinates(input), expected);
}
